part of geoclip;

class ComplexPolygon extends Polygon {
   /**
    * Only applies to the first [Polygon] and can only be used 
    * with a [Polygon] that contains one [Polygon]
    */
   List<Polygon> _List = new List<Polygon>();
   
   ComplexPolygon();
   
   ComplexPolygon.asHole(bool isHole) {
      _isHole = isHole;
   }
   
   bool operator==(Polygon other) {
     if (!(other is ComplexPolygon))
       return false;
     
     if (_isHole != other._isHole)
       return false;
     
     ComplexPolygon that = other as ComplexPolygon;

     if (_List.isNotEmpty && that._List.isNotEmpty) {
       Polygon thisp = _List[0];
       Polygon thatp = that._List[0];
       if (!(thisp == thatp)) return false;
     }
     
     // We should perform recusion because inside polygons could very
     // well be different.
//     for(Polygon p in _List) {
//       for(Polygon thatp in that._List) {
//         if (!(p == thatp))
//           return false;
//       }
//     }
     
     return true;
   }

   // TODO hashcode??
//   /**
//    * Return the hashCode of the object.
//    *
//    * @return an integer value that is the same for two objects
//    * whenever their internal representation is the same (equals() is true)
//    **/
//   int hashCode()
//   {
//      int result = 17 ;
//      result = 37*result + _List.hashCode();
//      return result;
//   }
   
//   /**
//    *
//    */
//   String toString()
//   {
//      return super.toString();
//   }
   
   /**
    * Resets [Polygon]
    */
   void clear() {
      _List.clear();
   }
   
   /**
    * Add [GPoint] to outer [Polygon]
    * If this is the first point added then a [ComplexPolygon] is
    * create as an inside [SimplePolygon] automatically.
    */
   void addCoord(double x, double y) {
     addPoint(new GPoint(x, y));
   }
   
   /**
    * Add [GPoint] to outer [Polygon]
    * If this is the first point added then a [ComplexPolygon] is
    * create as an inside [SimplePolygon] automatically.
    */
   void addPoint(GPoint p) {
      if (_List.isEmpty) {
         _List.add(new SimplePolygon());
      }
      _List[0].addPoint(p);
   }
   
   /**
    * Add an  [Polygon] to this [Polygon] - assumes that adding [Polygon] does not
    * have any  [Polygon]s.
    *
    * @throws GeoClipException if the number of inside [Polygon]s is greater than
    * zero and this [Polygon] was designated a hole. This would break the assumption
    * that only simple [Polygon]s can be holes.
    */
   void addPolygon(Polygon p) {
     if ((_List.isNotEmpty) && _isHole) {
       throw new GeoClipException("Cannot add polygon to a polygon marked as a Hole.");
     }
     _List.add(p);
   }
   
   /**
    * Return [true] if the polygon is empty
    */
   bool get isEmpty => _List.isEmpty;
   
   /**
    * Returns the bounding rectangle of this polygon.
    * <strong>WARNING</strong> Not supported on complex polygons.
    */
   GRectangle get bounds {
     if (_List.isEmpty) {
       return new GRectangle.zero();
     }
     else if (_List.length == 1) {
       Polygon ip = getPolygon(0);
       return ip.bounds;
     }
     else {
       throw new GeoClipException("bounds not supported on complex poly.");
     }
   }
   
   /**
    * Returns the polygon at this index.
    */
   Polygon getPolygon(int polyIndex) {
     return _List[polyIndex];
   }
   
   /**
    * Returns the number of inside polygons -  polygons are assumed to return one here.
    */
   int get numberOfPolygons => _List.length;
   
   /**
    * Return the number points of the first inside polygon
    */
   int get numberOfPoints => _List[0].numberOfPoints;
   
   /**
    * Return the point component of outer polygon.
    */
   GPoint getPoint(int index) {
     return _List[0].getPoint(index);
   }
   
   List<Polygon> get polygons => _List;
   
   /**
    * Return true if this polygon is a hole.  Holes are assumed to be inside polygons of
    * a more complex polygon.
    *
    * @throws [GeoClipException] if called on a [ComplexPolygon].
    */
   bool get isHole {
     if( _List.length > 1 ) {
       throw new GeoClipException( "Cannot call on a poly made up of more than one poly." );
     }
     return _isHole;
   }
   
   /**
    * Set whether or not this polygon is a hole.  Cannot be called on a complex polygon.
    *
    * @throws GeoClipException if called on a complex polygon.
    */
   void set isHole(bool isHole) {
     if ( _List.length > 1 ) {
       throw new GeoClipException( "Cannot call on a polygon made up of more than one polygon." );
     }
     _isHole = isHole;
   }
   
   bool get hasHoles {
     // If a single child polygon is marked as a hole then return true.
     bool hole = _isHole;
     
     if (!hole) {
       for(Polygon p in _List) {
         if (p is ComplexPolygon) {
           hole = p.hasHoles;
           if (hole)
             break;
         }
       }
     }
     
     return hole;
   }
   
   /**
    * Return true if the given inside polygon is contributing to the set operation.
    * This method should NOT be used outside the Clip algorithm.
    */
   bool isContributing(int polyIndex) {
     return _List[polyIndex].isContributing(0);
   }
   
   /**
    * Set whether or not this inside polygon is constributing to the set operation.
    * This method should NOT be used outside the Clip algorithm.
    *
    * @throws GeoClipException if called on a complex polygon
    */
   void setContributing(int polyIndex, bool contributes) {
     // TODO This causes issues with subtractions. How valid is this test?
//     if (_List.length != 1) {
//       throw new GeoClipException("Only applies to polygons of size 1");
//     }
     _List[polyIndex].setContributing(0, contributes);
   }
   
   /**
    * Return a Poly that is the intersection of this polygon with the given polygon.
    * The returned polygon could be complex.
    *
    * @return the returned Poly will be an instance of PolyDefault.
    */
   Polygon intersection(Polygon p) {
     return Boolean.intersection(p, this, "ComplexPolygon");
   }
   
   /**
    * Return a Poly that is the union of this polygon with the given polygon.
    * The returned polygon could be complex.
    *
    * @return the returned Poly will be an instance of PolyDefault.
    */
   Polygon add(Polygon p) {
     return Boolean.add(p, this, "ComplexPolygon");
   }
   
   /**
    * Return a Poly that is the exclusive-or of this polygon with the given polygon.
    * The returned polygon could be complex.
    *
    * @return the returned Poly will be an instance of PolyDefault.
    */
   Polygon xor(Polygon p) {
     return Boolean.xor(p, this, "ComplexPolygon");
   }
   
   /**
    * Return a Poly that is the difference of this polygon with the given polygon.
    * The returned polygon could be complex.
    *
    * @return the returned Poly will be an instance of PolyDefault.
    */
   Polygon subtract(Polygon p) {
     return Boolean.subtract(this, p, "ComplexPolygon");
   }
   
   /**
    * Return the area of the polygon in square units.
    */
   double get area {
      double area = 0.0 ;
      for( int i = 0 ; i < numberOfPolygons; i++ ) {
        Polygon p = getPolygon(i);
        double tarea = p.area * (p.isHole ? -1.0 : 1.0);
        area += tarea;
      }
      return area ;
   }

   @override
   String toString() {
     StringBuffer s = new StringBuffer();
     for( int i = 0 ; i < _List.length ; i++ ) {
       Polygon p = getPolygon(i);
       s.writeln("InnerPoly(${i}).hole=${p.isHole}");
       for( int j = 0 ; j < p.numberOfPoints; j++ ) {
         GPoint pt = p.getPoint(j);
         s.writeln(pt);
       }
     }
     return s.toString();
   }
   
}
