library caster_tree;

import 'dart:html';
import 'package:vector_math/vector_math.dart';

import '../utilities.dart';
import 'caster_strip.dart';
import 'caster_edge.dart';
import 'caster.dart';
import 'caster_tree_node.dart';
import 'light.dart';
import 'shadow_map.dart';

/*
The tree currently checks only Caster aaboxes.
The output are strips sent to the lighting code.

At each level we check if the caster's aabbox fits entirely.
If so we subdivide by 4 and look for the node the caster fits in.
If we can't find any child node then we stop and place the caster at the level
where the caster last fit entirely.
Of course a worst case is if the caster's bbox is at the center then it ends up
at the top level providing no benefit.

The tree has a max level, typically 5.

TODO:
A level is only added when a caster fits within it. This leads to a sparsely
populated tree.
 */
class CasterTree {
  static int MAX_DEPTH = 5;
  
  List<CasterStrip> _strips = new List<CasterStrip>();
  
  // This collection is used just for simple finding.
  List<Caster> _casters = new List<Caster>();
  
  List<Caster> _nodeCasters = new List<Caster>();
  List<Caster> _culledCasters = new List<Caster>();

  ShadowMap _map;
  CasterStrip _cs;
  CasterEdge _ce;
  bool _prevFacing;
  bool _facing;

  CasterTreeNode _nodes;
  
  String path;
  
  // BEGIN Debug only
  int count = 0;
  List<String> _colors = ["rgba(255,0,0, 0.25)",
                          "rgba(0,0,255, 0.25)",
                          "rgba(0,255,255, 0.25)",
                          "rgba(255,255,0, 0.25)",
                          "rgba(255,0,255, 0.25)",
                          "rgba(255,128,0, 0.25)",
                          "rgba(0,255,128, 0.25)",
                          "rgba(128,255,0, 0.25)",
                          "rgba(32,32,32, 0.25)",
                          "rgba(0,128,0, 0.25)"];
  int _colorIndex = 0;
  // END Debug only
  
  CasterTree(ShadowMap map) {
    _map = map;
    _nodes = new CasterTreeNode(this, null);
    _nodes.setMap(map);
  }
  
  /*
  This version uses a Quadtree to find qualifying casters. The casters are
  added to the tree based on their aabbox's.
  */
  List<CasterStrip> findStripsComplex(Light light) {
    _nodeCasters.clear();

    //count = 0;
    _nodes.findCasters(_nodeCasters, light.bounds);
    //print(count);
    
    // Finally, collect any casters that where parked at the root.
    List<Caster> rootCasters = _nodes.getCasters();
    if (rootCasters.isNotEmpty) {
      _nodeCasters.addAll(rootCasters);
    }

    //print(_nodeCasters.length);
    return findStripsIntermediateWithCasters(_nodeCasters, light);
  }
  
  List<CasterStrip> findStripsIntermediate(Light light) {
    return findStripsIntermediateWithCasters(_casters, light);
  }

  /*
  This version of findStrips is more complex in that it also checks aabboxes
  of edges.
  
  Check each edge for facing and range.
  If a facing changes then simply end the current strip and start a new one.
  if out-of-range then end the current strip and iterate to the next in-range
  edge.
      We start a New strip when:
      1) an edge is out of range of the light
      or
      2) and edge's facing attribute changes relative to a sibling edge.
  This version doesn't actually use the Quadtree for culling. See
  findStripsComplex()
  */
  List<CasterStrip> findStripsIntermediateWithCasters(List<Caster> casters, Light light) {
    _strips.clear();
    
    for(Caster caster in casters) {
      if (!caster.active)
        continue;
      
      // Cull Casters that are not intersecting light aabbox.
      Aabb2 casterBounds = caster.bounds;

      if (!casterBounds.intersectsWithAabb2(light.bounds))
        continue;

      List<CasterEdge> edgeList = caster.edgeList;

      // Check for trival case of a single edge.
      if (edgeList.length == 1) {
        CasterEdge ce = edgeList.first;
        bool inRange = edgeInRange(light.bounds, caster, ce);
        if (inRange) {
          _cs = new CasterStrip();
          _cs.start = ce;
          _cs.end = ce;
          _strips.add(_cs);
        }
        continue;
      }

      // This caster has 2 or more edges.
      _cs = null;

      // Find farthest edge as a starting point.
      _ce = edgeList.first;
      if (caster.closed) {
        _ce = findFarthestEdgeFromLight(light.position, edgeList);
      }

      CasterEdge startEdge = _ce;
      
      // First we need to establish where to start a strip. This occurs
      // when we hit the first edge in range relative to the current
      // edge.
      // When we wrap around to the "start" we closeout any strip in
      // progress and exit.
      // Once we establish a starting edge we create a new strip
      // and begin marching until we hit:
      // 1) The Start Edge
      //    closeout any strip in progress.
      // 2) Edge out of range
      //    closeout any strip in progress and begin looking for
      //    a new starting Edge. If we reach Start then exit.
      // 3) Inflection change.
      //    Closeout current strip and begin looking for a new starting
      //    edge. If we reach Start then exit.
      //
      // For "open" casters we need to watch out for the end/null edge.
      
      // true = trying to establish a strip.
      // false = strip in progress
      bool establishing = true;
      bool stripInProgress = false;
      CasterEdge lastEdge;
      
      do {
        if (establishing) {
          // We are trying to establish a strip which means we are looking
          // for an edge in range.
          bool inRange = edgeInRange(light.bounds, caster, _ce);
          if (!inRange) {
            _ce = _ce.nextEdge;
            continue;
          }
          
          // We found a Edge in range. begin new strip.
          _cs = new CasterStrip();
          _cs.start = _ce;
          _cs.end = _ce;
          establishing = false;
          stripInProgress = true;
        }
        
        // Strip in progress. Is inflection going to change on the next edge.
        if (_ce.nextEdge != null) {
          bool inflectionChange = isInflectionChangeComingUp(light.position, _ce);
          if (inflectionChange) {
            // The next edge will cause an inflection change.
            // closeout current strip.
            _cs.end = _ce;
            _strips.add(_cs);
            stripInProgress = false;
            
            // The next edge will cause an inflection change but is it in
            // range? We don't know so we attempt to reestablish a Start edge.
            establishing = true;
          }
          else {
            // The next edge isn't an inflection change. However, it
            // could be out of range.
            // If the next edge is out of range then we closeout and
            // begin marching again.
            bool inRange = edgeInRange(light.bounds, caster, _ce.nextEdge);
            if (inRange) {
              // The Next edge is in range. update end marker and continue.
              _cs.end = _ce;
            }
            else {
              // the Next edge is out of range. closeout any strip in progress
              // and attempt to reestablish a Start edge.
              _cs.end = _ce;
              _strips.add(_cs);
              stripInProgress = false;
              
              establishing = true;
            }
          }
        }
        
        // Move to the next edge. However, if the next is null--which can
        // be for "open" casters--then we capture the trailing edge to
        // potentially closeout the last strip in progress.
        if (_ce.nextEdge == null)
          lastEdge = _ce;
        
        _ce = _ce.nextEdge;
        
        // The null check is for "open" casters.
      } while (_ce != startEdge && _ce != null);
      
      if (stripInProgress) {
        // The intent here is to properly cap-off the strip with the
        // caster's trailing edge.
        // If the _ce=null then the caster was "open" and _ce "ran off"
        // then end so we use the "lastEdge" instead.
        // Otherwise, _ce has actually wrapped around--for an "closed"
        // caster which means we simply backup to the previous edge.
        if (_ce != null)
          _cs.end = _ce.prevEdge;
        else
          _cs.end = lastEdge;
        _strips.add(_cs);
      }
      
    } // End caster loop
    
    //print(_strips.length);
    return _strips;
  }
  
  CasterEdge findEdgeInRange(List<CasterEdge> edgeList, Aabb2 lightShadowBounds, Aabb2 casterShadowBounds, Caster caster, CasterEdge start, CasterEdge ce) {
      CasterEdge e = null;
//      CasterEdge end = edgeList[edgeList.length - 1];
      
      bool inRange = edgeInRange(lightShadowBounds, caster, ce);
      
      while (!inRange && ce != start) {
        ce = ce.nextEdge;
        if (ce == null) // handles "open" casters
          break;
        inRange = edgeInRange(lightShadowBounds, caster, ce);
        if (inRange) {
          break;
        }
      }
      
      if (inRange) {
        e = ce;
      }
      return e;
  }
  
  /**
  0 = next edge is valid.
  1 = next edge is out of range
  2 = next edge changed facing
  3 = next edge is Start edge. indicates wrapped around.
  */
  int checkNextEdge(List<CasterEdge> edgeList, Aabb2 lightBounds, Caster caster, Vector2 lightPos, CasterEdge start, CasterEdge ce) {
    CasterEdge ne = ce.nextEdge;

    if (ne == start || ne == null)
      return 3;
    
    bool inRange = edgeInRange(lightBounds, caster, ne);
    
    if (!inRange) {
      return 1;
    }

    _facing = ne.isEdgeFacingLight(lightPos);

    if (_prevFacing != _facing) {
      _prevFacing = _facing;
      return 2;
    }
    
    _prevFacing = _facing;
    _ce = ne;
    
    return 0;
  }
  
  bool isInflectionChangeComingUp(Vector2 lightPos, CasterEdge ce) {
    bool facing = ce.isEdgeFacingLight(lightPos);
    bool nextFacing = ce.nextEdge.isEdgeFacingLight(lightPos);
    return facing != nextFacing;
  }
  
  bool edgeInRange(Aabb2 lightBounds, Caster caster, CasterEdge ce) {
    bool intersect = lightBounds.intersectsWithAabb2(ce.bounds);

    return intersect;
  }
  
  /*
  This version of findStrips is simple. It checks only for light-caster
  intersections. It scans the whole caster from start for end.
  */
  List<CasterStrip> findStripsSimple(Light light) {
    _strips.clear();
    
    // We start a New strip when:
    // 1) an edge is out of range of the light
    // or
    // 2) and edge's facing attribute changes relative to the previous edge.
    for(Caster caster in _casters) {
      if (!caster.active)
          continue;
      
      // Cull Casters that are not intersecting light aabbox.
      Aabb2 casterBounds = caster.bounds;

      if (!casterBounds.intersectsWithAabb2(light.bounds))
          continue;

      CasterStrip cs = null;

      List<CasterEdge> edgeList = caster.edgeList;

      // Check for trival case of a single edge.
      if (edgeList.length == 1) {
        CasterEdge ce = edgeList.first;
        cs = new CasterStrip();
        cs.start = ce;
        cs.end = ce;
        _strips.add(cs);
        continue;
      }
      
      cs = new CasterStrip();
      
      // Because GeoClip isn't numerically robust we can't rely on two edges
      // that are end-wise coincident with each other otherwise you get slivers
      // at the joint. To solve this we need to pick a better starting 
      // edge rather than defaulting to "start". So we locate an edge
      // with a vertex farthest from the light and start there.
      // Note: This approach doesn't work if the light goes inside the
      // closed polygon.
      
      Vector2 lightPos = light.position;
      
      CasterEdge startEdge = edgeList.first;
      if (caster.closed) {
        startEdge = findFarthestEdgeFromLight(lightPos, edgeList);
      }

      bool prevFacing = startEdge.isEdgeFacingLight(lightPos);
      cs.start = startEdge;
      
      CasterEdge currentEdge = startEdge.nextEdge;
      
      do {
        bool facing = currentEdge.isEdgeFacingLight(lightPos);
        if (facing == prevFacing) {
          // Inflection hasn't changed. continue;
          cs.end = currentEdge;
          cs.edges++;
        }
        else {
          // Inflection changed. End current strip.
          cs.end = currentEdge.prevEdge;
          _strips.add(cs);

          // Start New strip.
          cs = new CasterStrip();
          cs.start = currentEdge;
          cs.end = currentEdge;
        }

        currentEdge = currentEdge.nextEdge;
        prevFacing = facing;
        // We also check for currentEdge=null for "open" polygons.
      } while (currentEdge != startEdge && currentEdge != null);

      _strips.add(cs);
    }
    
    return _strips;
  }
  
  CasterEdge findFarthestEdgeFromLight(Vector2 lightPos, List<CasterEdge> edgeList) {
    CasterEdge startEdge = edgeList.first;
    double pd = double.NEGATIVE_INFINITY;
    
    for(CasterEdge edge in edgeList) {
      double d = distanceBetweenByDouble(lightPos.x, lightPos.y, edge.v0.x, edge.v0.y);
      if (d > pd) {
        startEdge = edge;
        pd = d;
      }
    }
    
    return startEdge;
  }
  
  List<Caster> findCasters(Light light) {
    _culledCasters.clear();
    
    // TODO use tree and NOT list.
    for(Caster caster in _casters) {
      if (!caster.active)
          continue;
      
      Aabb2 casterBounds = caster.bounds;

      if (casterBounds.intersectsWithAabb2(light.bounds)) {
        _culledCasters.add(caster);
      }
    }
    
    return _culledCasters;
  }
  
  void addCaster(Caster caster) {
    // If using findStripsSimple then uncomment to populate simple collection
    _casters.add(caster);
    // otherwise use insertCaster.
    insertCaster(caster);
  }
  
  void insertCaster(Caster caster) {
    path = "";
    Aabb2 mapBounds = _map.bounds;
    
    if (!_nodes.isInitialized())
        _nodes.init(mapBounds, 0);
    
    if (!caster.active) // debug
        return;
    
    Aabb2 casterBounds = caster.bounds;
    _nodes.addCaster(caster, casterBounds);
  }
  
  void updateCaster(Caster caster) {
    // Remove caster if it is already in the tree.
    // For now we simply traverse.
    bool emptyOccurred = removeCaster(caster);
    insertCaster(caster);

    if (emptyOccurred) {
      path = "";
      int count = _nodes.clean();
      if (count == 0) {
        //System.out.println("uninit: " + this);
        _nodes.unInit();
      }
    }
  }
  
  bool removeCaster(Caster caster) {
    path = "";
    //_casters.remove(caster);
    // Iterate through tree to find node where caster is located.
    bool emptyOccurred = _nodes.removeCaster(caster);
    
    return emptyOccurred;
  }
  
  void removeAllCasters() {
    _nodes.removeAllCasters();
  }
  
  void draw(CanvasRenderingContext2D context) {
    _nodes.draw(context);
    
    _colorIndex = 0;
    for(CasterStrip cs in _strips) {
      cs.draw(context, _colors[_colorIndex]);
      _colorIndex = (_colorIndex + 1) % _colors.length;
    }
    //_strips.forEach((CasterStrip cs) => cs.draw(context, _colors[_colorIndex]));
  }
}
