library caster_tree_node;
import 'dart:html';

import 'caster_tree.dart';
import 'caster.dart';
import 'shadow_map.dart';

import 'package:vector_math/vector_math.dart';

/*
|---------|---------|
|         |         |
|    2    |    1    |
|         |         |
|---------|---------|
|         |         |
|    3    |    4    |
|         |         |
|---------|---------|
*/
class CasterTreeNode {
  final List<CasterTreeNode> _nodes = new List<CasterTreeNode>(4);
  final CasterTreeNode _parent;
  final CasterTree _tree;
  
  int id;
  int lvl;
  
  final List<Caster> _casters = new List<Caster>();
  Aabb2 _aabbox = new Aabb2();
  bool _initialized = false;
  
  // debug stuff
  Aabb2 quad1r = new Aabb2();
  Aabb2 quad2r = new Aabb2();
  Aabb2 quad3r = new Aabb2();
  Aabb2 quad4r = new Aabb2();

  ShadowMap _map;
  
  bool _hasCasters = false;
  
  CasterTreeNode(this._tree, this._parent);
  
  void setMap(ShadowMap map) {
    _map = map;
  }
  
  List<Caster> getCasters() {
      return _casters;
  }
  
  bool hasCasters() {
      return _hasCasters;
  }
  
  void init(Aabb2 parentBounds, int level) {
    _initialized = true;
    lvl = level;
    setBoundsMinMax(parentBounds.min, parentBounds.max);
    
    // Add quads to tree
    double width = parentBounds.max.x - parentBounds.min.x;
    double height = parentBounds.max.y - parentBounds.min.y;
    double hw = width / 2.0;
    double hh = height / 2.0;
    
    CasterTreeNode quad1 = new CasterTreeNode(_tree, this);
    quad1.setMap(_map);
    quad1.id = 1;
    quad1.lvl = level + 1;
    quad1.setBounds(parentBounds.min.x + hw, parentBounds.min.y, hw, hh);
    _nodes[0] = quad1;
    quad1r = new Aabb2.minMax(quad1._aabbox.min, quad1._aabbox.max);

    CasterTreeNode quad2 = new CasterTreeNode(_tree, this);
    quad2.setMap(_map);
    quad2.id = 2;
    quad2.lvl = level + 1;
    quad2.setBounds(parentBounds.min.x, parentBounds.min.y, hw, hh);
    _nodes[1] = quad2;
    quad2r = new Aabb2.minMax(quad2._aabbox.min, quad2._aabbox.max);

    CasterTreeNode quad3 = new CasterTreeNode(_tree, this);
    quad3.setMap(_map);
    quad3.id = 3;
    quad3.lvl = level + 1;
    quad3.setBounds(parentBounds.min.x, parentBounds.min.y + hh, hw, hh);
    _nodes[2] = quad3;
    quad3r = new Aabb2.minMax(quad3._aabbox.min, quad3._aabbox.max);

    CasterTreeNode quad4 = new CasterTreeNode(_tree, this);
    quad4.setMap(_map);
    quad4.id = 4;
    quad4.lvl = level + 1;
    quad4.setBounds(parentBounds.min.x + hw, parentBounds.min.y + hh, hw, hh);
    _nodes[3] = quad4;
    quad4r = new Aabb2.minMax(quad4._aabbox.min, quad4._aabbox.max);
  }
  
  bool unInit() {
    bool castersPresent = false;

    for(CasterTreeNode n in _nodes) {
      if (n != null) {
        if (n._hasCasters) {
          castersPresent = true;
          break;
        }
        else {
          castersPresent = n.unInit();
          if (castersPresent)
            break;
        }
      }
    }

    if (!castersPresent) {
      //System.out.println("no casters present. uninitializing: " + this);
      _nodes[0] = null;
      _nodes[1] = null;
      _nodes[2] = null;
      _nodes[3] = null;
      _initialized = false;
    }
    else {
        //System.out.println("Casters present. CAN'T uninitializing: " + this);
    }

    return castersPresent;
  }
  
  bool isInitialized() {
    return _initialized;
  }
  
  bool fits(Aabb2 bounds) {
    return _aabbox.containsAabb2(bounds);
  }
  
  bool touches(Aabb2 bounds) {
    return _aabbox.intersectsWithAabb2(bounds);
  }
  
  void addCaster(Caster caster, Aabb2 casterBounds) {
    //_tree.path += String.valueOf(lvl) + ":" + String.valueOf(id) + ",";
    //System.out.println("entered level: "+lvl + ", quad: "+id);
    if (lvl > CasterTree.MAX_DEPTH) {
      //System.out.println("MAXED: adding " + caster + ", l: "+ lvl + ", quad: "+ id);
      //System.out.println("MAXED: " + _tree.path);
      _casters.add(caster);
      _hasCasters = true;
      return;
    }
    
    // Check each child for a fit. If it fits then walk the child.
    // If the caster doesn't fit any of the children then add the caster
    // to this node.
    bool oneFit = false;
    
    if (!isInitialized())
      init(_aabbox, lvl);
    
    //int nlvl = level + 1;
    
    // If the caster doesn't fit any of the child nodes AND none of them
    // currently have no casters then we can unInit this node.
    int casterCount = 0;
    for(CasterTreeNode n in _nodes) {
      //System.out.println("checking n: " + n +" | "+n._aabbox);
      //System.out.println("casterBounds: "+casterBounds);
      if (n._hasCasters)
        casterCount++;
      if (n.fits(casterBounds)) {
        //print("fits into: $n");
        n.addCaster(caster, casterBounds);
        //_hasCasters = n.hasCasters();
        oneFit = true;
        break;
      }
    }

    if (!oneFit) {
      if (casterCount == 0)
        unInit();
      //System.out.println("adding " + caster + ", l: "+ lvl + ", quad: "+ id);
      //System.out.println("path: " + _tree.path);
      //_map.setPath(_tree.path);
      _casters.add(caster);
      _hasCasters = true;
    }
  }
  
  bool removeCaster(Caster caster) {
    bool casterRemoved = false;
    
    if (_casters.contains(caster)) {
      //System.out.println("removing Caster: " + caster + " from: "+this+". size: " + _casters.size());
      casterRemoved = _casters.remove(caster);
      if (_casters.isEmpty) {
        //System.out.println("Caster: " + caster + " removed. size: " + _casters.size());
        _hasCasters = false;
      }
    }
        
    if (!casterRemoved) {
      // The caster wasn't removed from this node. Iterate into the
      // children.
      for(CasterTreeNode n in _nodes) {
        casterRemoved = n.removeCaster(caster);
        if (casterRemoved) {
          break;
        }
      }
    }
    
    return casterRemoved;
  }

  void removeAllCasters() {
//        for(CasterTreeNode n : _nodes) {
//            if (n.hasCasters()) {
//                n.clear();
//                n.removeAllCasters();
//            }
//        }
  }

  // This method scans the tree looking for nodes that have no casters in
  // "any" of the children nodes along a path.
  // As long as there are children in one of the quads then node can't be
  // removed/UnInit.
  // - The simplest case is that this node has no casters and no children in
  // which case this node can be removed by the caller.
  int clean() {
    // Traverse each node checking for nodes that have caster children.
    // If the count stays at 0 then this node's nodes have no casters
    // and we can eliminate this node.
    //_tree.path += String.valueOf(lvl) + ":" + String.valueOf(id) + ",";
    
    int count = _casters.length;
    //System.out.println("cleaning: " + this);
    
    // At level 0 there are 4 quads (1,2,3,4)
    for(CasterTreeNode n in _nodes) {
//            System.out.println("clean into: " + n);
      count += n.clean();
//            System.out.println("clean ["+n +"] count: " + count);
      if (count == 0) {
//                System.out.println("child uninit: " + n);
        n.unInit();
      }
      else {
//                System.out.println("children inside of: " + n);
          //System.out.println("path: " + _tree.path);
          //break;
      }
    }

    if (count == 0 && _nodes.isNotEmpty) {
//            System.out.println("uninit: " + this);
//            System.out.println("path: " + _tree.path);
      unInit();
    }
    
//        System.out.println("exit clean: " + this);
    
    return count;
  }
  
  void clear() {
    _casters.clear();
  }
  
  // the dimensions are in shadow-space.
  void setBounds(double minX, double minY, double width, double height) {
    _aabbox = new Aabb2.minMax(new Vector2(minX, minY), new Vector2(minX + width, minY + height));
  }
  
  void setBoundsMinMax(Vector2 min, Vector2 max) {
    _aabbox = new Aabb2.minMax(min, max);
  }
  
  /*
  If the light's bounds intersect a node then collect all the children of
  that node.
  */
  void findCasters(List<Caster> casters, Aabb2 lightBounds) {
    //print("working at [${_tree.count}]:" + toString());
    _tree.count++;
    
    // Collect any casters in children.
    for(CasterTreeNode n in _nodes) {
      if (n != null) {
        //print("checking: $n");
        if (n.touches(lightBounds)) {
          // collect casters of this node.
          if (n._casters.isNotEmpty) {
            //System.out.println("collecting: "+ n);
            casters.addAll(n._casters);
          }
          // Iterate downward into child for other casters
          n.findCasters(casters, lightBounds);
        }
      }
    }
  }
  
  void draw(CanvasRenderingContext2D context) {
    for(CasterTreeNode n in _nodes) {
      if (n != null) {
        n.draw(context);
      }
    }
    
    _drawQuad(quad1r, context);
    _drawQuad(quad2r, context);
    _drawQuad(quad3r, context);
    _drawQuad(quad4r, context);
  }

  void _drawQuad(Aabb2 quad, CanvasRenderingContext2D context) {
    context..save()
      ..lineWidth = 1
      ..strokeStyle = "rgba(0,0,0, 0.75)"
      ..beginPath()
      ..moveTo(quad.min.x, quad.min.y)
      ..lineTo(quad.min.x, quad.max.y)
      ..lineTo(quad.max.x, quad.max.y)
      ..lineTo(quad.max.x, quad.min.y)
      ..closePath()
      ..stroke()
      ..restore();
  }
  
  @override
  String toString() {
//    StringBuilder s = new StringBuilder();
//    s.append("lvl:"+lvl+", quad: "+id+"| casters: " + _casters.size() + ", hascasters: " + _hasCasters);
//
//    return s.toString();
    return "lvl: $lvl, quad: $id, casters: ${_casters.length}, hasCasters: $_hasCasters";
  }
}
