library caster_light;

import 'package:vector_math/vector_math.dart';
import 'dart:math';
import 'dart:html';

import 'caster_tree.dart';
import 'caster_strip.dart';
import 'shadow_map.dart';
import 'polygon.dart';
import '../utilities.dart';

/*
 * Light has a visual handle center.
 * a visual polygon ring (also for computation.)
 * a resultant lit area.
 * The light is dragged using mouse events.
 */
class Light extends Polygon {
  List<Vector2> wVertices = new List<Vector2>();

  double radius = 150.0;
  double selectRadius;

  int segments = 16;

  double _theta_inc;

  Map _map;
  ShadowMap _shadowMap;
  List<CasterStrip> _casterStrips;
  
  Light.fromJson(ShadowMap shadowMap, Map map) {
    _map = map;
    _shadowMap = shadowMap;
    
    num v = _map["Radius"] as num;
    radius = v.toDouble();
    selectRadius = radius / 10.0;

    active = _map["Active"] as bool;
    segments = _map["Segments"] as int;
    
    Map pos = _map["LightPos"];
    v = pos["x"] as num;
    position.x = v.toDouble();
    v = pos["y"] as num;
    position.y = v.toDouble();
    
    _buildRim();
    
    calcBounds(wVertices);
  }

  List<CasterStrip> get strips => _casterStrips;

  bool isPointOnCenter(Point location) {
    double distance = distanceBetweenByDouble(location.x.toDouble(), location.y.toDouble(), position.x, position.y);
    return distance < selectRadius;
  }

  @override
  void begin() {
    super.begin();
    wVertices.clear();
  }

  void moveBy(double dx, double dy) {
    position.setValues(position.x + dx, position.y + dy);
    
    // Recalc world-space coords
    _buildRim();
    calcBounds(wVertices);

    dirty = true;
  }

  void _buildRim() {
    begin();
    _theta_inc = 2.0 * PI / segments;
    double theta = 2.0 * PI;

    // Note: Canvas +Y axis is oriented downwards.
    // So for CCW we decrement theta starting from 2PI.
    for(int i = 0; i < segments; i++) {
      double x = radius * cos(theta);
      double y = radius * sin(theta);
      vertices.add(new Vector2(x, y));
      wVertices.add(new Vector2(x + position.x, y + position.y));
      theta -= _theta_inc;
    }

    closed = true;
  }

  void reCast() {
    CasterTree tree = _shadowMap.tree;
    
//    _casterStrips = tree.findStripsSimple(this);
    _casterStrips = tree.findStripsComplex(this);
  }
  
  @override
  void draw(CanvasRenderingContext2D context) {
    if (!active)
      return;
    
    context.save();
    // Draw light rim before clipping
    if (vertices.isNotEmpty) {
      context..lineWidth = 1
             ..strokeStyle = "rgba(255,128,0,0.5)";

      Vector2 p = vertices[0];
      // By adding position we map from local to world space.
      context..beginPath()
             ..moveTo(p.x + position.x, p.y + position.y);

      Iterable<Vector2> subList = vertices.getRange(1, vertices.length);
      for(Vector2 p in subList) {
        context.lineTo(p.x + position.x, p.y + position.y);
      }
      context..closePath()
             ..stroke();
    }

    // Draw inner drag circle
    context..beginPath()
           ..arc(position.x, position.y, radius / 10, 0, 2 * PI, false)
           ..fillStyle = 'rgba(255,255,0, 0.25)'
           ..fill()
           ..lineWidth = 1
           ..strokeStyle = 'rgba(0,127,0, 1.0)'
           ..stroke();

    context.restore();
  }

}

