library caster_edge_path;

import 'dart:math';
import 'dart:html';

import 'package:geoclip/geoclip.dart' as Geo;
import 'package:vector_math/vector_math.dart';
import 'light.dart';
import 'caster_strip.dart';
import 'caster_edge.dart';

class LightEdgePath {
  final Light _light;
  final Map _map;
  
  double selectRadius;
  double _theta_inc;

  final Vector2 _vector = new Vector2.zero();
  Vector2 _lightRayEndPoint = new Vector2.zero();

  Vector2 _ray00 = new Vector2.zero();
  Vector2 _ray0 = new Vector2.zero();
  Vector2 _ray1 = new Vector2.zero();
  Vector2 _sum = new Vector2.zero();

  // _litArea is the final result.
  Geo.Polygon _litArea;
  
  // For each strip we create a shadow polygon that is subtracted
  // from a start light polygon.
  Geo.Polygon _shadow = new Geo.ComplexPolygon();
  
  LightEdgePath(this._map, this._light) {
    selectRadius = _light.radius / 10.0;
    _theta_inc = 2.0 * PI / _light.segments;
  }
  
  Geo.Polygon get litArea => _litArea;
  
  void build(List<CasterStrip> casterStrips) {
    // For now convert light vertices into Geo's polygon format.
    _litArea = new Geo.ComplexPolygon();
    _light.wVertices.forEach((Vector2 v) => _litArea.addCoord(v.x, v.y));

    if (casterStrips != null) {
      for(CasterStrip strip in casterStrips) {
        _shadow.clear();
        createShadowStrip(strip);
        // Subtract from light
        try {
          _litArea = _litArea.subtract(_shadow);
        }
        on Exception catch(e) {
          print("LightEdgePath.build: $e");
        }
      }
    }

  }

  Geo.Polygon createShadowStrip(CasterStrip strip) {
      CasterEdge currentEdge = strip.start;

      // ----------------------------------------------------
      // First add inner edge.
      // ----------------------------------------------------
      _shadow.addCoord(currentEdge.v0.x, currentEdge.v0.y);

      while (currentEdge != strip.end) {
        _shadow.addCoord(currentEdge.v1.x, currentEdge.v1.y);
       
        currentEdge = currentEdge.nextEdge;
      }

      // Now process the end edge.
      _shadow.addCoord(currentEdge.v1.x, currentEdge.v1.y);

      // ----------------------------------------------------
      // Now add outer edge.
      // ----------------------------------------------------
      currentEdge = strip.end;
      
      while (currentEdge != strip.start) {
        Vector2 lightRayEndPoint0 = formLightRayEndPoint(_light.position, currentEdge.v0.x, currentEdge.v0.y, _light.radius * 1.5);
        Vector2 lightRayEndPoint1 = formLightRayEndPoint(_light.position, currentEdge.v1.x, currentEdge.v1.y, _light.radius * 1.5);
        
        _shadow.addCoord(lightRayEndPoint1.x, lightRayEndPoint1.y);
        
        _ray00.setFrom(lightRayEndPoint0);
        _ray0.setFrom(lightRayEndPoint0.sub(_light.position));
        _ray1.setFrom(lightRayEndPoint1.sub(_light.position));
        _sum.setFrom(_ray0.add(_ray1).normalize().scale(_light.radius * 1.5).add(_light.position));
        _shadow.addCoord(_sum.x, _sum.y);

        currentEdge = currentEdge.prevEdge;
      }

      // Now process the start edge.
      Vector2 lightRayEndPoint0 = formLightRayEndPoint(_light.position, currentEdge.v0.x, currentEdge.v0.y, _light.radius * 1.5);
      Vector2 lightRayEndPoint1 = formLightRayEndPoint(_light.position, currentEdge.v1.x, currentEdge.v1.y, _light.radius * 1.5);

      _shadow.addCoord(lightRayEndPoint1.x, lightRayEndPoint1.y);

      _ray00.setFrom(lightRayEndPoint0);
      _ray0.setFrom(lightRayEndPoint0.sub(_light.position));
      _ray1.setFrom(lightRayEndPoint1.sub(_light.position));
      _sum.setFrom(_ray0.add(_ray1).normalize().scale(_light.radius * 1.5).add(_light.position));
      _shadow.addCoord(_sum.x, _sum.y);

      _shadow.addCoord(_ray00.x, _ray00.y);

      return _shadow;
  }

//  Geo.Polygon createAShadow(CasterEdge edge) {
//    Point2D lightRayEndPoint0 = formLightRayEndPoint(lightPos, edge.v0.x, edge.v0.y, _radius * 1.5);
//    Point2D lightRayEndPoint1 = formLightRayEndPoint(lightPos, edge.v1.x, edge.v1.y, _radius * 1.5);
//
//    Point2D v0 = lightRayEndPoint0.subtract(lightPos);
//    Point2D v1 = lightRayEndPoint1.subtract(lightPos);
//    Point2D sum = v0.add(v1).normalize().multiply(_radius * 1.5).add(lightPos);
//
//    Path2D.Double path = new Path2D.Double();
//    path.moveTo(edge.v0.x, edge.v0.y);
//    path.lineTo(edge.v1.x, edge.v1.y);
//    path.lineTo(lightRayEndPoint1.getX(), lightRayEndPoint1.getY());
//    path.lineTo(sum.getX(), sum.getY());
//    path.lineTo(lightRayEndPoint0.getX(), lightRayEndPoint0.getY());
//    path.closePath();
//    
//    return new Area(path);
//  }
  
  Vector2 formLightRayEndPoint(Vector2 lightPos, double x, double y, double length) {
    Vector2 p = new Vector2(x - lightPos.x, y - lightPos.y);
    p.normalize().scale(length).add(lightPos);
    return p;
  }

  void draw(CanvasRenderingContext2D context) {
    context.save();
    
    if (_litArea != null) {
      drawPolygon(_litArea, "0,255,0,0.176", "0,90,0,1.0", context);
      //drawPolygon(_shadow, "255,255,0,0.117", "255,0,0,0.176", context);    
    }
    
//    context..lineWidth = 1
//        ..strokeStyle = "rgb(255,0,255)"
//        ..beginPath()
//        ..moveTo(_light.position.x, _light.position.y)
//        ..lineTo(_ray00.x, _ray00.y)
//        ..stroke();
//    
//    context..strokeStyle = "rgb(0,0,255)"
//        ..beginPath()
//        ..moveTo(_light.position.x, _light.position.y)
//        ..lineTo(_ray1.x + _light.position.x, _ray1.y + _light.position.y)
//        ..stroke();
//    
//    context..strokeStyle = "rgb(255,0,0)"
//        ..beginPath()
//        ..moveTo(_light.position.x, _light.position.y)
//        ..lineTo(_sum.x, _sum.y)
//        ..stroke();
    
    context.restore();
  }
  
  void drawSimplePolygon(Geo.SimplePolygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
    if (polygon.isEmpty)
      return;
    
    List<Geo.GPoint> points = polygon.points;
    
    context..lineWidth = 1
        ..strokeStyle = "rgba(" + strokeColor + ")"
        ..fillStyle = "rgba(" + fillColor + ")"
        ..beginPath()
        ..moveTo(points[0].x, points[0].y);
        
    Iterable<Geo.GPoint> subList = points.getRange(1, points.length);
    subList.forEach((Geo.GPoint p) => context.lineTo(p.x, p.y));
    context..closePath()
        ..fill()
        ..stroke();
  }

  void drawPolygon(Geo.Polygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
    if (polygon is Geo.SimplePolygon)
      drawSimplePolygon(polygon, fillColor, strokeColor, context);
    else
      drawComplexPolygon(polygon, fillColor, strokeColor, context);
  }

  void drawComplexPolygon(Geo.ComplexPolygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
    // Complex polygons can have Simple polygons that are actually a part
    // of the Complex which makes them holes.
    // So we ask the Complex if it has any Simples acting as holes. If so
    // then we render differently.
    List<Geo.Polygon> polygons = polygon.polygons;
    
    if (polygon.hasHoles) {
      polygons.forEach((Geo.Polygon gp) => drawComplexPolygonWithHoles(gp, fillColor, strokeColor, context));
    }
    else {
      polygons.forEach((Geo.Polygon gp) => drawPolygon(gp, fillColor, strokeColor, context));
    }
    
  }

  void drawComplexPolygonWithHoles(Geo.ComplexPolygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
    // This renderer only considers a simple case where SimplePolygons represent holes.
    // Other more complex arrangements will need a better renderer.
    if (polygon.isEmpty)
      return;

    context..lineWidth = 1
        ..strokeStyle = "rgba(" + strokeColor + ")"
        ..fillStyle = "rgba(" + fillColor + ")"
        ..beginPath();
    
    _drawComplexPolygonWithHoles(polygon, fillColor, strokeColor, context);
    
    context..fill()
        ..stroke();
  }

  void _drawComplexPolygonWithHoles(Geo.ComplexPolygon polygon, String fillColor, String strokeColor, CanvasRenderingContext2D context) {
    for(Geo.Polygon ply in polygon.polygons) {
      if (ply is Geo.ComplexPolygon) {
        _drawComplexPolygonWithHoles(ply, fillColor, strokeColor, context);
      }
      else {
        Geo.SimplePolygon sp = ply as Geo.SimplePolygon;
        
        List<Geo.GPoint> points = sp.points;
        
        context.moveTo(points[0].x, points[0].y);
        
        Iterable<Geo.GPoint> subList = points.getRange(1, points.length);
        subList.forEach((Geo.GPoint p) => context.lineTo(p.x, p.y));
        
        context.closePath();
      }
    }
  }

}
