library shadow_map;

import 'package:vector_math/vector_math.dart';
import 'dart:html';

import 'light.dart';
import 'caster.dart';
import 'caster_tree.dart';
import 'light_edge_path.dart';
import '../utilities.dart';

class ShadowMap {
  Aabb2 bounds;

  List<Light> _lights = new List<Light>();
  List<Caster> _casters = new List<Caster>();
  Map _map;
  
  Light _selectedLight;
  bool mouseOnLight = false;
  double previousPointX = -1.0;
  double previousPointY = -1.0;
  bool mouseDown = false;

  CasterTree _quadTree;
  LightEdgePath _litArea;
  
  Vector2 _pt = new Vector2.zero();
  
  ShadowMap(Map map, double width, double height) {
    _map = map;
    bounds = new Aabb2.minMax(new Vector2.zero(), new Vector2(width, height));

    _quadTree = new CasterTree(this);
    
    List<Map> lights = map["Lights"];
    
    for(Map m in lights) {
      Light l = new Light.fromJson(this, m);
      _lights.add(l);
    }

    _selectedLight = _lights[0];
    
    _litArea = new LightEdgePath(_map, _selectedLight);
    
    List<Map> casters = map["Casters"];

    for(Map m in casters) {
      Caster c = new Caster.fromJson(m);
      _casters.add(c);
      
      // Load caster into quad tree
      if (c.active)
        _quadTree.addCaster(c);
    }
    
  }
  
  CasterTree get tree => _quadTree;
  Light get selectedLight => _selectedLight;
  
  void touchDown(Point p) {
    _selectedLight = selectLight(p);
    if (_selectedLight != null) {
      previousPointX = p.x.toDouble();
      previousPointY = p.y.toDouble();
      mouseOnLight = true;
      reCast();
    }
    mouseDown = true;
  }

  void touchUp(Point p) {
    mouseDown = false;
    mouseOnLight = false;
  }

  void touchMove(Point p) {
    if (mouseDown && mouseOnLight) {
      double dx = p.x.toDouble() - previousPointX;
      double dy = p.y.toDouble() - previousPointY;
      
      // Check that this move doesn't put the light inside one of the
      // casters.
      List<Caster> casters = tree.findCasters(_selectedLight);
      bool inside = false;
      
      _pt.setValues(_selectedLight.position.x + dx, _selectedLight.position.y + dy);
      
      for(Caster c in casters) {
        if (c.closed) {
          inside = polygonContainsPoint(c.wVertices, _pt);
          if (inside)
            break;
        }
      }
      
      if (!inside) {
        _selectedLight.moveBy(dx, dy);
        reCast();
      }
    }
    
    previousPointX = p.x.toDouble();
    previousPointY = p.y.toDouble();
  }

  void reCast() {
    // TODO The light at this position causes an exception [866.0,640.0]
    if (_selectedLight != null) {
      _selectedLight.reCast();
      _litArea.build(_selectedLight.strips);
    }
  }
  
  Light selectLight(Point p) {
    Light lt;
    for(Light l in _lights) {
      if (l.isPointOnCenter(p)) {
        lt = l;
        break;
      }
    }
    return lt;
  }
  
  void draw(CanvasRenderingContext2D context) {
    _casters.forEach((Caster c) => c.draw(context));
    _lights.forEach((Light l) => l.draw(context));
    _quadTree.draw(context);
    
    if (_litArea != null) {
      _litArea.draw(context);
    }
  }
}